from __future__ import absolute_import

from pytest_xdist_tracker.tracker import TestRunner, TestTracker


def pytest_addoption(parser):
    """
    Command line options for our plugin
    Parameters
    ----------
    parser: _pytest.config.Parser
    """
    group = parser.getgroup("xdist_tracker")
    group.addoption(
        "--xdist-stats",
        action="store",
        default="xdist_stats",
        dest="xdist_stats",
        help=(
            "File pattern to save tests which were run on the xdist node (by default %(default)s). "
            "As result if will run with `-n2` will generated 2 artefacts like:"
            "xdist_stats_worker_gw0.txt and xdist_stats_worker_gw1.txt"
        ),
    )
    group.addoption(
        "--from-xdist-stats",
        action="store",
        default=None,
        dest="from_xdist_stats",
        help=(
            "File (generated by `--xdist-stats`) with tests(nodeid) to run in single thread, "
            "could be helpful to reproduce issues "
            "related to coupled tests which corrupted or doesn't clear some state after self "
            "(not work with xdist `-n`)"
        ),
    )


def pytest_configure(config):
    """
    Enable this reporter when tests run with XDIST
    """
    is_run_with_xdist = bool(
        config.pluginmanager.get_plugin("xdist") and config.option.numprocesses
    )
    is_run_xdist_worker = bool(getattr(config, "workerinput", None))
    is_run_to_reproduce = bool(config.getoption("--from-xdist-stats"))
    if (is_run_with_xdist or is_run_xdist_worker) and not is_run_to_reproduce:
        reporter = TestTracker(config)
        config.pluginmanager.register(reporter, name="xdist_tracker")
    if is_run_to_reproduce and not (is_run_with_xdist or is_run_xdist_worker):
        runner = TestRunner(config)
        config.pluginmanager.register(runner, name="xdist_runner")



import pytest


@pytest.fixture
def default_test_nodeid():
    return "tests/test_xxx.py::test_xxx"


@pytest.fixture
def file_name():
    # can be overwritten via local fixture
    return "xdist_stats_worker_gw2.txt"


@pytest.fixture
def expected_file_path(tmpdir, file_name):
    return str(tmpdir / file_name)


@pytest.fixture
def expected_file(expected_file_path, default_test_nodeid):
    with open(expected_file_path, "w") as f:
        f.write(default_test_nodeid)
    return expected_file_path



from __future__ import absolute_import

import io

import pytest
from six.moves import urllib_parse

ENCODING = "utf-8"


def is_xdist_worker(config):
    """
    Parameters
    ----------
    config: _pytest.config.Config

    Return
    ------
    bool
        `True` if this is an xdist worker, `False` otherwise
    """
    return hasattr(config, "workerinput")


def get_xdist_worker_id(config):
    """
    Parameters
    ----------
    config: _pytest.config.Config

    Returns
    -------
    str
        the id of the current worker ('gw0', 'gw1', etc) or 'master'
        if running on the 'master' node.

        If not distributing tests (for example passing `-n0` or not passing `-n` at all)
        also return 'master'.
    """
    worker_input = getattr(config, "workerinput", None)
    if worker_input:
        return worker_input["workerid"]
    return "master"


class TestTracker(object):
    """
    Plugin track tests which run in particular xdist node
    As result save artifact with with these tests
    In case when have some flaky test it could be helpful to reproduce it
    """

    def __init__(self, config):
        self.config = config
        self.is_loadfile = self.config.getoption("dist") == "loadfile"
        self.storage = []

    def get_name(self, item):
        """
        Usually returns full test case name
            "tests/path/test_module.py::TestCase::test_one"
        But if was passed dist=loadfile will be enough just test module name
            "tests/path/test_module.py"

        Parameters
        -----------
        item : _pytest.main.Item

        Returns
        -------
        str
        """
        if self.is_loadfile:
            # TODO #7 dist in node always `no`
            # ("test_module_path.py", line number, "test_case")
            return item.location[0]
        return urllib_parse.quote(item.nodeid)

    def add(self, item):
        """
        Parameters
        -----------
        item : _pytest.main.Item
        """
        name = self.get_name(item)
        if name not in self.storage:
            self.storage.append(name)

    @property
    def file_path(self):
        """
        Making path base on passed patter and worker id

        Returns
        -------
        str
            "xdist_stats_worker_gw1.txt"
            "xdist_stats_worker_gw2.txt"
            ...
        """
        worker_id = get_xdist_worker_id(self.config)
        file_name = "{}_worker_{}.txt".format(
            self.config.getoption("--xdist-stats"), worker_id
        )
        return str(self.config.rootdir / file_name)

    def store(self):
        """
        Save as artifact all tests which were run inside particular xdist node
        tests separate by new line
        """

        with io.open(self.file_path, "wb") as file:
            content = "\n".join(self.storage)
            file.write(content.encode(ENCODING))

    @pytest.hookimpl(hookwrapper=True, trylast=True)
    def pytest_sessionfinish(self):
        """
        Storing all test modules which were executed on this xdist node
        Sometimes needs to debugging coupled tests when failure happens
        This list of test modules sorted as they were executed,
        then can help to reproduce an issue
        """
        if is_xdist_worker(self.config):
            self.store()
        yield

    @pytest.hookimpl(hookwrapper=True, trylast=True)
    def pytest_runtest_call(self, item):
        """
        Parameters
        ----------
        item : _pytest.main.Item
        """
        self.add(item)
        yield


class TestRunner(object):
    """
    This plugin help to run particular tests from artifact which was generated via `TestTracker`
    """

    def __init__(self, config):
        """
        Parameters
        ----------
        config: _pytest.config.Config
        """
        self._target_tests = None
        self.config = config
        # patch of passed arguments `tests/...` to reduce collection runtime
        self.config.args[:] = self.target_test_modules

    def read_target_tests(self):
        """
        Reads artifact which was generated via `TestRunTracker` plugin

        Returns
        -------
        List[str]
            [
                "tests/backend/test_one.py::test_one",
                "tests/backend/test_one.py::test_two",
                ...
            ]
        """
        file_path = self.config.getoption("--from-xdist-stats")
        test_cases = []
        with io.open(file_path, "rb") as file:
            for line in file:
                test_cases.append(
                    urllib_parse.unquote(line.decode(ENCODING).rstrip("\n"))
                )
        return test_cases

    @property
    def target_tests(self):
        """
        Returns
        -------
        List[str]
            [
                "tests/backend/test_one.py::test_one",
                "tests/backend/test_one.py::test_two",
                ...
            ]
        """
        if self._target_tests is None:
            self._target_tests = self.read_target_tests()
        return self._target_tests

    @property
    def target_test_modules(self):
        """
        Returns
        -------
        set[str]
        """
        return list({test_case.split("::")[0] for test_case in self.target_tests})

    def find_necessary(self, items):
        """
        Parameters
        ----------
        items: List[_pytest.main.Item]

        Returns
        -------
        Generator[pytest.Item]
        """
        return (item for item in items if item.nodeid in self.target_tests)

    def sorted_as_target_tests(self, items):
        """
        Parameters
        ----------
        items: List[_pytest.main.Item]

        Returns
        -------
        Generator
        """
        return sorted(items, key=lambda x: self.target_tests.index(x.nodeid))

    @pytest.hookimpl(hookwrapper=True)
    def pytest_collection_modifyitems(self, items):
        """
        Parameters
        ----------
        items : List[pytest.Item]
        """
        items[:] = self.sorted_as_target_tests(self.find_necessary(items))
        yield



import pytest


@pytest.fixture
def target_tests(testdir):
    testdir.makepyfile(
        """
            import pytest

            def test_fail0():
                assert 0

            def test_fail1():
                raise ValueError()

            def test_ok():
                pass

            @pytest.mark.skip
            def test_skip():
                assert 0
        """
    )
    return testdir


@pytest.mark.parametrize("numprocesses", (1, 3))
def test_storing_worker_artifacts(target_tests, numprocesses):
    report = target_tests.runpytest("-n", str(numprocesses))
    result = report.parseoutcomes()
    assert result["passed"] == 1
    assert result["skipped"] == 1
    assert result["failed"] == 2
    for n in range(numprocesses):
        generated_artifact = target_tests.tmpdir.join(
            "xdist_stats_worker_gw{}.txt".format(n)
        )
        assert generated_artifact.isfile()


def test_not_run_tracker_without_xdist(target_tests):
    report = target_tests.runpytest("-n0", "-s")
    result = report.parseoutcomes()
    assert result["passed"] == 1
    assert result["skipped"] == 1
    assert result["failed"] == 2
    generated_artifact = target_tests.tmpdir.join("xdist_stats_worker_gw0.txt")
    assert not generated_artifact.isfile()


def test_run_tests_from_artifact(target_tests):
    """
    test_run_tests_from_artifact.py::test_fail0
    test_run_tests_from_artifact.py::test_ok
    test_run_tests_from_artifact.py::test_fail1
    """
    lines = [
        "test_run_tests_from_artifact.py::test_ok",
        "test_run_tests_from_artifact.py::test_fail1",
    ]
    f = target_tests.maketxtfile("\n".join(lines))
    report = target_tests.runpytest("--from-xdist-stats", str(f))
    result = report.parseoutcomes()
    assert result["passed"] == 1
    assert result["failed"] == 1


def test_run_tests_not_from_artifact(target_tests):
    """
    test_run_tests_not_from_artifact.py::test_fail0
    test_run_tests_not_from_artifact.py::test_ok
    test_run_tests_not_from_artifact.py::test_fail1
    """
    lines = [
        "test_run_tests_not_from_artifact.py::test_ok",
        "test_run_tests_not_from_artifact.py::test_fail1",
    ]
    f = target_tests.maketxtfile("\n".join(lines))
    report = target_tests.runpytest("--from-xdist-stats", str(f), "-n1")
    result = report.parseoutcomes()
    assert result["passed"] == 1
    assert result["skipped"] == 1
    assert result["failed"] == 2



import os

try:
    from unittest import mock
except ImportError:
    from mock import mock

import pytest
from _pytest.config import Config
from six.moves import urllib_parse

from pytest_xdist_tracker.tracker import TestRunner as Runner
from pytest_xdist_tracker.tracker import TestTracker as Tracker

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError  # python2


def create_pytest_test_item(name):
    node = mock.create_autospec(pytest.Item, spec_set=True)
    node.nodeid = u"tests/backend/unit/test_awesome.py::test_one_{}".format(name)
    return node


@pytest.fixture
def config(tmpdir):
    c = mock.create_autospec(Config)
    c.getoption.return_value = None
    c.workerinput = {"workerid": "gw2"}
    c.rootdir = tmpdir
    c.args = ["tests/backend/unit"]
    return c


@pytest.fixture
def node():
    return create_pytest_test_item(0)


class TestTracker(object):
    FILE_NAME = "my_super_name"

    @pytest.fixture
    def file_name(self):
        return "{}_worker_gw2.txt".format(self.FILE_NAME)

    @pytest.fixture
    def tracker(self, config):
        config.getoption.return_value = self.FILE_NAME
        return Tracker(config=config)

    def test_instance(self, tracker, config):
        assert tracker.config == config
        assert tracker.storage == []

    def test_add(self, tracker, node):
        tracker.add(node)
        assert tracker.storage == [urllib_parse.quote(node.nodeid)]

    def test_add_multiply_times(self, tracker, node):
        tracker.add(node)
        tracker.add(node)
        tracker.add(node)
        assert len(tracker.storage) == 1
        assert tracker.storage == [urllib_parse.quote(node.nodeid)]

    def test_file_path(self, tracker, expected_file_path):
        assert tracker.file_path.endswith("{}_worker_gw2.txt".format(self.FILE_NAME))
        assert tracker.file_path == expected_file_path

    def test_store(self, tracker, node, expected_file_path):
        tracker.add(node)
        tracker.store()
        assert os.path.isfile(expected_file_path), "File not exist"
        with open(expected_file_path) as file_content:
            content = file_content.read().strip()
        assert urllib_parse.unquote(content) == node.nodeid

    def test_store_with_file_pattern(self, tracker, node, expected_file_path):
        tracker.add(node)
        tracker.store()
        assert os.path.isfile(expected_file_path), "File not exist"
        with open(expected_file_path) as file_content:
            content = file_content.read().strip()
        assert urllib_parse.unquote(content) == node.nodeid

    def test_store_empty(self, tracker, expected_file_path):
        tracker.store()
        assert os.path.isfile(expected_file_path), "File not exist"
        with open(expected_file_path) as file_content:
            content = file_content.read().strip()
        assert content == ""

    def test_pytest_sessionfinish(self, expected_file_path, tracker):
        next(tracker.pytest_sessionfinish())
        assert os.path.isfile(expected_file_path), "File not exist"

    def test_pytest_sessionfinish_for_master_node(self, tracker):
        delattr(tracker.config, "workerinput")
        next(tracker.pytest_sessionfinish())
        assert not os.path.isfile(tracker.file_path), "File exists"

    def test_pytest_runtest_call(self, node, tracker):
        next(tracker.pytest_runtest_call(node))
        assert tracker.storage == [urllib_parse.quote(node.nodeid)]


class TestRunner(object):
    FILE_NAME = "my_awesome_name"

    @pytest.fixture
    def file_name(self):
        return "{}_worker_gw2.txt".format(self.FILE_NAME)

    @pytest.fixture
    def runner(self, config, expected_file):
        config.getoption.return_value = expected_file
        assert config.args == ["tests/backend/unit"]
        r = Runner(config=config)
        assert config.args != ["tests/backend/unit"] and config.args
        return r

    @pytest.fixture
    def target_items(self, node):
        """
        Emulate failed test nodes
        Parameters
        ----------
        node: pytest.Item
            "tests/backend/unit/test_awesome.py::test_one_0"

        Returns
        -------
        List[pytest.Item]
            [
                "tests/backend/unit/test_awesome.py::test_one_0",
                "tests/backend/unit/test_awesome.py::test_one_12",
                "tests/backend/unit/test_awesome.py::test_one_11",
            ]
        """
        return [node, create_pytest_test_item(12), create_pytest_test_item(11)]

    @pytest.fixture
    def items(self, node, target_items):
        """
        Returns
        -------
        List[str]
            [
                "tests/backend/unit/test_awesome.py::test_one_1",
                "tests/backend/unit/test_awesome.py::test_one_2",
                "tests/backend/unit/test_awesome.py::test_one_3",
                "tests/backend/unit/test_awesome.py::test_one_4",
                "tests/backend/unit/test_awesome.py::test_one_5",
                "tests/backend/unit/test_awesome.py::test_one_6",
                "tests/backend/unit/test_awesome.py::test_one_7",
                "tests/backend/unit/test_awesome.py::test_one_8",
                "tests/backend/unit/test_awesome.py::test_one_9",
                "tests/backend/unit/test_awesome.py::test_one_10",
                "tests/backend/unit/test_awesome.py::test_one_0",
                "tests/backend/unit/test_awesome.py::test_one_12",
                "tests/backend/unit/test_awesome.py::test_one_11",
            ]
        """
        items = []
        for idx in range(1, 11):
            items.append(create_pytest_test_item(idx))
        # emulates shuffling of tests (diff between run with xdist and without)
        items.extend(target_items)
        return items

    @pytest.fixture
    def target_tests(self, target_items):
        """
        Parameters
        ----------
        List[pytest.Item]
            [
                "tests/backend/unit/test_awesome.py::test_one_0",
                "tests/backend/unit/test_awesome.py::test_one_12",
                "tests/backend/unit/test_awesome.py::test_one_11",
            ]

        Returns
        -------
        List[str]
            [
                "tests/backend/unit/test_awesome.py::test_one_0",
                "tests/backend/unit/test_awesome.py::test_one_11",
                "tests/backend/unit/test_awesome.py::test_one_12",
            ]
        """
        return sorted([node.nodeid for node in target_items])

    def test_instance(self, runner, config, default_test_nodeid):
        assert runner.config == config
        assert runner._target_tests is not None
        assert runner.target_tests == runner._target_tests
        assert runner.target_tests == [default_test_nodeid]

    def test_read_target_tests(self, expected_file_path, runner, default_test_nodeid):
        tests = runner.read_target_tests()
        assert tests == [default_test_nodeid]

    def test_read_target_tests_file_absent(self, runner):
        runner.config.getoption.side_effect = "file_not_exist.txt"
        with pytest.raises(FileNotFoundError):
            runner.read_target_tests()

    def test_target_tests(self, runner, node, default_test_nodeid):
        assert runner.target_tests == [default_test_nodeid]

    def test_find_necessary(self, runner, target_tests, items):
        runner._target_tests = target_tests
        assert sorted(i.nodeid for i in runner.find_necessary(items)) == sorted(
            target_tests
        )

    def test_sorted_as_target_tests(self, runner, target_tests, items):
        runner._target_tests = target_tests
        target_items = runner.find_necessary(items)
        assert [
            i.nodeid for i in runner.sorted_as_target_tests(target_items)
        ] == target_tests

    def test_pytest_collection_modifyitems(self, runner, items, target_tests):
        runner._target_tests = target_tests
        assert len(items) != len(target_tests)
        next(runner.pytest_collection_modifyitems(items))
        assert len(items) == len(target_tests), "Was not filter target test items"
        assert [i.nodeid for i in items] == target_tests



import codecs

from setuptools import find_packages, setup

VERSION_FILE = "pytest_xdist_tracker/_version.py"


with codecs.open("README.md", "r", "utf-8") as fh:
    long_description = fh.read()


setup(
    name="pytest-xdist-tracker",
    use_scm_version={
        "write_to": VERSION_FILE,
        "local_scheme": "dirty-tag",
    },
    setup_requires=["setuptools_scm==5.0.2"],
    author="Denis Korytkin",
    author_email="dkorytkin@gmail.comm",
    description="pytest plugin helps to reproduce failures for particular xdist node",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DKorytkin/pytest-xdist-tracker",
    keywords=["py.test", "pytest", "xdist plugin", "tracker", "failed tests"],
    py_modules=[
        "pytest_xdist_tracker.plugin",
        "pytest_xdist_tracker.tracker",
    ],
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "pytest>=3.5.1",
        "pytest-xdist>=1.23.2",
        "pytest-forked>=1.0.1",
        "six>=1.15.0",
    ],
    entry_points={"pytest11": ["tracker = pytest_xdist_tracker.plugin"]},
    license="MIT license",
    python_requires=">=2.7",
    classifiers=[
        "Framework :: Pytest",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Topic :: Utilities",
    ],
)



import pytest


def test_execute_xdist_tracker_plugin(testdir):
    config = testdir.parseconfigure("-n1")
    assert config.option.xdist_stats == "xdist_stats"
    assert config.pluginmanager.hasplugin("xdist_tracker")
    assert not config.pluginmanager.hasplugin("xdist_runner")


def test_execute_xdist_tracker_plugin_with_custom_options(testdir):
    config = testdir.parseconfigure("-n2", "--xdist-stats", "xXx")
    assert config.option.xdist_stats == "xXx"
    assert config.pluginmanager.hasplugin("xdist_tracker")
    assert not config.pluginmanager.hasplugin("xdist_runner")


def test_execute_xdist_tracker_plugin_without_xdist(testdir):
    config = testdir.parseconfigure("--xdist-stats", "xXx")
    assert config.option.xdist_stats == "xXx"
    assert not config.pluginmanager.hasplugin("xdist_tracker")
    assert not config.pluginmanager.hasplugin("xdist_runner")


def test_execute_xdist_runner(testdir, expected_file):
    config = testdir.parseconfigure("--from-xdist-stats", expected_file)
    assert config.option.from_xdist_stats == expected_file
    assert not config.pluginmanager.hasplugin("xdist_tracker")
    assert config.pluginmanager.hasplugin("xdist_runner")


def test_execute_xdist_runner_with_xdist(testdir, expected_file):
    config = testdir.parseconfigure("--from-xdist-stats", expected_file, "-n 2")
    assert config.option.from_xdist_stats == expected_file
    assert not config.pluginmanager.hasplugin("xdist_tracker")
    assert not config.pluginmanager.hasplugin("xdist_runner")


def test_execute_both_plugins(testdir, expected_file):
    config = testdir.parseconfigure(
        "--from-xdist-stats", expected_file, "-n 2", "--xdist-stats", "x2"
    )
    assert config.option.from_xdist_stats == expected_file
    assert config.option.xdist_stats == "x2"
    assert not config.pluginmanager.hasplugin("xdist_tracker")
    assert not config.pluginmanager.hasplugin("xdist_runner")
